@@ -1,10 +1,10 @@
 /* Invoke unistd-like functions, but avoid some glitches.
 
-   Copyright (C) 2001, 2003, 2005, 2009-2021 Free Software Foundation, Inc.
+   Copyright (C) 2001, 2003, 2005, 2009-2024 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 3 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -17,6 +17,11 @@
 
 /* Written by Paul Eggert and Eric Blake.  */
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+
 int dup_safer (int);
 int fd_safer (int);
 int pipe_safer (int[2]);
@@ -29,3 +34,8 @@ int fd_safer_flag (int, int);
 #if GNULIB_PIPE2_SAFER
 int pipe2_safer (int[2], int);
 #endif
+
+
+#ifdef __cplusplus
+}
+#endif
