@@ -47,10 +47,10 @@ class HTTPSServer(StoppableHTTPServer):
                                                os.getenv('srcdir', '.'),
                                                'certs',
                                                'server-key.pem'))
-        self.socket = ssl.wrap_socket(
+        ctx = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS_SERVER)
+        ctx.load_cert_chain(CERTFILE, KEYFILE)
+        self.socket = ctx.wrap_socket(
             sock=socket.socket(self.address_family, self.socket_type),
-            certfile=CERTFILE,
-            keyfile=KEYFILE,
             server_side=True
         )
         self.server_bind()
@@ -68,7 +68,7 @@ class _Handler(BaseHTTPRequestHandler):
     def get_rule_list(self, name):
         return self.rules.get(name)
 
-    # The defailt protocol version of the server we run is HTTP/1.1 not
+    # The default protocol version of the server we run is HTTP/1.1 not
     # HTTP/1.0 which is the default with the http.server module.
     protocol_version = 'HTTP/1.1'
 
